clear all; close all; clear classes; format short; format compact;

console=1;

plot_figs=1;   %Switch on/off plotting 
plot_emb=1;    %Switch on/off plotting of the embedding 


prefiks='Minimal-Depth';


%______________________________________________________________
% GENERATE A GRAPH


rngseed=3;   
%rand('state',0);                  %resets the generator to its initial state.
rand('state',rngseed);             %integer J resets the generator to its J-th state.
%rand('state',sum(100*clock));	   %different seed each time

num_of_nodes=10; squaresize=20; connthr=8; titlestring='10 nodes, thr=8, 20x20';
[adj_mat coords] = f_gen_wireless_graph(num_of_nodes,squaresize,connthr);

if plot_figs==1,
   figsize=10; figborder=0.2; figure; whitebg('white'); hold on; title(titlestring);
   axis([-figsize-figborder figsize+figborder -figsize-figborder figsize+figborder]);
   axis square; %axis off; %axis equal;

   edgewidth=0.5;edgecolor=[1 .5 .5];
   f_graphplot_edges(adj_mat,coords,edgewidth,edgecolor)

   nodesize=10; nodecolor='k'; fontsize=10; xoffset=0.04; yoffset=0.04;
   f_graphplot_nodes(adj_mat,coords,nodesize,nodecolor,fontsize,xoffset,yoffset)
end; %if plot_figs

%______________________________________________________________ 
% FIND A TREE 

root=3;
fprintf(2,'finding a minimal depth tree, root=%d\n',root);
[distances_from_root discover_times tree]=f_mindepth_tree(adj_mat,root);

titlestring='Spanning Tree';

if plot_figs==1,
   % PLOT HIGHLIGHT TREE
   %figsize=10; figborder=0.2; figure; whitebg('white'); hold on;
   %axis([-figsize-figborder figsize+figborder -figsize-figborder figsize+figborder]);
   %axis square; %axis off; %axis equal;
   
   title([prefiks ' ' titlestring ' highlighted; root=' num2str(root)]);
   
   edgewidth=0.5;edgecolor='k';xoffset=0.04; yoffset=0.04; fontsize=10;
   f_treeplot_highlight(adj_mat,coords,tree,edgewidth,edgecolor,xoffset,yoffset,fontsize)

   nodesize=10; nodecolor='k'; fontsize=10; xoffset=0.04; yoffset=0.04;
   f_graphplot_nodes(adj_mat,coords,nodesize,nodecolor,fontsize,xoffset,yoffset)
   

   % PLOT TOPOLOGY TREE
   figure; hold on; title([prefiks ' ' titlestring ' - topology']);
   xoffset=0.02; yoffset=0.02; fontsize=10;
   labels=1:length(adj_mat);
   f_treeplot_topology(tree,xoffset,yoffset,fontsize,labels);
   
end; %if plot_figs

%______________________________________________________________
% "discover time" of the root is 0, and so on. nodes will be embedded in the same order. 
% 'indeks' stores the names of the nodes in the order in which they will be embedded

[sorted indeks]=sort(discover_times);


%______________________________________________________________
% EMBED AND PLOT

%________________
% INITIALIZE

alpha1(root)=0;   alpha2(root)=2*pi;   alpha1_init(root)=alpha1(root);
aa(root)=exp(i*alpha1(root));
bb(root)=exp(i*alpha2(root));
geo_ctr(root)=2/( conj( aa(root) + bb(root) ) );   geo_rad(root)=abs(geo_ctr(root) - aa(root));
loc(root)=-0.1-i*0.1;


dokade=num_of_nodes;  %The last node to be embedded. Used for debugging. Normally all nodes should be embedded.

%______________________________________________________________
% THIS IS THE ONLINE EMBEDDING ALGORITHM FROM THE PAPER
% FINDS THE COORDINATES OF THE NODES AND STORES THEM IN 'LOC' 

for ii=2:dokade,
   curnode=indeks(ii);   parent=tree(curnode);

%  1) OWNED SPACE
   if alpha1(parent)~=alpha1_init(parent),
     alpha1(curnode)=alpha1(parent);
     alpha2(curnode)=( alpha1(parent)+alpha2(parent) ) / 2;
   else
     alpha1(curnode)=( alpha1(parent)+alpha2(parent) ) / 2;
     alpha2(curnode)=( alpha1(curnode)+alpha2(parent) ) / 2;
   end %if
   alpha1_init(curnode)=alpha1(curnode);
   alpha1(parent)=alpha2(curnode);

%  2) CENTER AND RADIUS OF GEODESIC THAT DEFINES THE OWNED SPACE OF CURNODE
   aa(curnode)=exp(i*alpha1(curnode));
   bb(curnode)=exp(i*alpha2(curnode));
   geo_ctr(curnode)=2 / ( conj( aa(curnode) + bb(curnode)) );
   geo_rad(curnode)=abs(geo_ctr(curnode)-aa(curnode));

%  3) HYPERBOLIC COORDINATE OF CURNODE
   loc(curnode)=geo_rad(curnode)^2/(conj(loc(parent)-geo_ctr(curnode)))+geo_ctr(curnode);

end



%______________________________________________________________
if plot_emb==1,

   %OPEN FIGURE FOR POINCARE DISK
      edgewidth=0.5;
      figure; hold on; figsize=1; figborder=0.02; title([prefiks ' ' titlestring ' - embedded']);
      axis([-figsize figsize -figsize figborder]); axis equal; axis off;
      %unit circle;
         theta = (512:1024)/1024*2*pi; x = cos(theta); y = sin(theta); plot(x,y,'-k','linewidth',edgewidth);
      %x y axes
         plot([-figsize,figsize],[0,0],'-k','linewidth',edgewidth);  %x oska
         plot([0,0],[-figsize,figborder],'-k','linewidth',edgewidth);  %y oska


   
   dotsize=7;  fontsize=7;

   

   %PLOT ROOT
   curnode=root;
   plot(real(loc(curnode)),imag(loc(curnode)),'.k','markersize',dotsize);
   set(text(real(loc(curnode))+0.005,imag(loc(curnode))-0.005,num2str(curnode)),'FontSize',fontsize);

   %PLOT OTHER NODES, THEIR GEODESICS AND EDGES TO THEIR PARENTS
   for ii=2:dokade,

      %PLOT NODE
      curnode=indeks(ii);  parent=tree(curnode);
      plot(real(loc(curnode)),imag(loc(curnode)),'.k','markersize',dotsize);
      set(text(real(loc(curnode))-0.007,imag(loc(curnode))-0.03,num2str(curnode)),'FontSize',fontsize);

      %PLOT GEODESIC. need to find the part of the geodesic to plot and deal with 2pi jump problem
      a1=angle(aa(curnode)-geo_ctr(curnode));  a2=angle(bb(curnode)-geo_ctr(curnode));
      if real(geo_ctr(curnode))>0,
        if (a1>0) & (a2<0),a2=a2+2*pi; elseif (a1<0) & (a2>0), a1=a1+2*pi; end
        step=abs(a1-a2)/500;   theta = min(a1,a2):step:max(a1,a2);
      else
        step=-abs(a1-a2)/500;  theta = max(a1,a2):step:min(a1,a2);
      end
      zgeo = geo_ctr(curnode)+geo_rad(curnode)*exp(i*theta);
      plot(zgeo,'-b','linewidth',edgewidth);

      %PLOT EDGE
      xa=real(loc(parent))+1.1e-8; ya=imag(loc(parent))+1e-8;
      xb=real(loc(curnode)); yb=imag(loc(curnode));
      aaa=-(xb-xa)/(yb-ya); bbb=(xa^2-xb^2+(ya-yb)*(ya+yb))/(2*(ya-yb));
      %geodesic circle center and radius
        xc=(xa^2*yb-xb^2*ya+(ya-yb)*(ya*yb-1))/(2*(xa*yb-xb*ya));  yc=aaa*xc+bbb;  R=sqrt((yc-ya)^2+(xc-xa)^2);
      a1=angle(loc(parent)-xc-i*yc);  a2=angle(loc(curnode)-xc-i*yc);
      if xc>0,
        if     (a1>0) & (a2<0), a2=a2+2*pi;
        elseif (a1<0) & (a2>0), a1=a1+2*pi; end
        step=abs(a1-a2)/500;    theta = min(a1,a2):step:max(a1,a2);
      else
        step=-abs(a1-a2)/500;   theta = max(a1,a2):step:min(a1,a2);
      end
      xuc = R*cos(theta)+xc; yuc = R*sin(theta)+yc;
      plot(xuc,yuc,'-k','linewidth',edgewidth);

   end
   
end; %if plot_figs





