#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <tmwtypes.h>
#include <gmp.h>
#include <mpfr.h>

#define PRECIZNOST 10000U
#define BUFFER 42000U


void c_mult(mpfr_ptr e, mpfr_ptr f, const mpfr_ptr a, const mpfr_ptr b, const mpfr_ptr c, const mpfr_ptr d) {
    //  CALCULATES e+fi = (a+bi)(c+di) =ac +adi +bci -bd

    mpfr_t t;
    mpfr_init2 (t, PRECIZNOST);

    //  *e=a*c-b*d;
    mpfr_mul (e, a, c, MPFR_RNDN);
    mpfr_mul (t, b, d, MPFR_RNDN);
    mpfr_sub (e, e, t, MPFR_RNDN);

    //  *f=a*d+b*c;
    mpfr_mul (f, a, d, MPFR_RNDN);
    mpfr_mul (t, b, c, MPFR_RNDN);
    mpfr_add (f, f, t, MPFR_RNDN);

    mpfr_clear(t);
}



void c_div(mpfr_ptr e, mpfr_ptr f, const mpfr_ptr a, const mpfr_ptr b, const mpfr_ptr c, const mpfr_ptr d) {

//  CALCULATES e+fi = (a+bi)/(c+di)
//  DENOMINATOR  (c+di) (c-di) = cc +dd -cdi +cdi = c^2 + d^2

    mpfr_t t1, t2;
    mpfr_inits2 (PRECIZNOST, t1, t2, (mpfr_ptr) 0);

//  double tmp=c*c+d*d;
    mpfr_sqr(t1,c,MPFR_RNDN);
    mpfr_sqr(t2,d,MPFR_RNDN);
    mpfr_add(t1,t1,t2,MPFR_RNDN);

//  *e=(a*c+b*d)/tmp;
    mpfr_mul (e, a, c, MPFR_RNDN);
    mpfr_mul (t2, b, d, MPFR_RNDN);
    mpfr_add (e, e, t2, MPFR_RNDN);
    mpfr_div (e, e, t1, MPFR_RNDN);

//  *f=(b*c-a*d)/tmp;
    mpfr_mul (t2, b, c, MPFR_RNDN);
    mpfr_mul (f, a, d, MPFR_RNDN);
    mpfr_sub (f, t2, f, MPFR_RNDN);
    mpfr_div (f, f, t1, MPFR_RNDN);

    mpfr_clears (t1, t2, (mpfr_ptr) 0);
}




void hiprast(const mpfr_ptr Z1r, const mpfr_ptr Z1i, const mpfr_ptr Z2r, const mpfr_ptr Z2i, mpfr_ptr izlez) {

    mpfr_t t1, t2, t3, eden, t4;
    mpfr_inits2 (PRECIZNOST, t1, t2, t3, eden, t4, (mpfr_ptr) 0);


    /*   HYPERBOLIC DISTANCE
     *                Zr{sources(ii)})  Z1r
     *                Zi{sources(ii)})  Z1i
     *                Zr{dests(ii)})    Z2r
     *                Zi{dests(ii)})    Z2i
     *
     *
     *
     *                IMPLEMENTS
     *
     *                double(  2*       atanh (  abs(Z1-Z2)  ./  abs( 1-  Z1 .* conj(Z2)  )     )      );
     *
     *                double(  2*       atanh (  abs( Z1r+i*Z1i-Z2r-i*Z2i )  ./  abs( 1-  Z1 .* conj(Z2)  )     )      );
     *
     */

    mpfr_sub(t1,Z1r,Z2r,MPFR_RNDN);
    mpfr_sub(t2,Z1i,Z2i,MPFR_RNDN);

    mpfr_hypot(t3,t1,t2,MPFR_RNDN);

    MPFR_SIGN(Z2i)=(MPFR_SIGN(Z2i) < 0) ? 1 : -1;
  
    c_mult(t1, t2, Z1r, Z1i, Z2r, Z2i);

  
    MPFR_SIGN(t2)=(MPFR_SIGN(t2) < 0) ? 1 : -1;

    mpfr_set_d (eden, 1.0, MPFR_RNDN);
    mpfr_sub(t1, eden, t1, MPFR_RNDN);


    mpfr_hypot(t4,t1,t2,MPFR_RNDN);



    mpfr_div (t3, t3, t4, MPFR_RNDN);

    mpfr_atanh (t3, t3, MPFR_RNDN);

    mpfr_mul_ui (izlez, t3, (unsigned long) 2,  MPFR_RNDN);



    mpfr_clears (t1,t2, t3, eden, t4, (mpfr_ptr) 0);

    MPFR_SIGN(Z2i)=(MPFR_SIGN(Z2i) < 0) ? 1 : -1;
}




int glavna(const int_T nn, const int_T *predecessors, const int_T root, const int_T *redosled, int_T *skoklist_grabl) {




        unsigned long i;
        int vrateno=0;


        //  ________ INITIALIZE ARRAYS OF  nn+1 MPFR COORDINATES  ____________________________________
        mpfr_t *Zr; Zr=(mpfr_t *)malloc((1+nn)*sizeof(mpfr_t));
        mpfr_t *Zi; Zi=(mpfr_t *)malloc((1+nn)*sizeof(mpfr_t));
        mpfr_t *alpha; alpha=(mpfr_t *)malloc((1+nn)*sizeof(mpfr_t));
        mpfr_t *beta;  beta= (mpfr_t *)malloc((1+nn)*sizeof(mpfr_t));
        mpfr_t *geo_rad; geo_rad=(mpfr_t *)malloc((1+nn)*sizeof(mpfr_t));

        mpfr_t *geo_ctr_r; geo_ctr_r=(mpfr_t *)malloc((1+nn)*sizeof(mpfr_t));
        mpfr_t *geo_ctr_i; geo_ctr_i=(mpfr_t *)malloc((1+nn)*sizeof(mpfr_t));
        mpfr_t *aa_r;      aa_r=(mpfr_t *)malloc((1+nn)*sizeof(mpfr_t));
        mpfr_t *aa_i;      aa_i=(mpfr_t *)malloc((1+nn)*sizeof(mpfr_t));
        mpfr_t *bb_r;      bb_r=(mpfr_t *)malloc((1+nn)*sizeof(mpfr_t));
        mpfr_t *bb_i;      bb_i=(mpfr_t *)malloc((1+nn)*sizeof(mpfr_t));



        for(i=1;i<=nn;i++) {
              mpfr_init2(Zr[i],PRECIZNOST);
              mpfr_init2(Zi[i],PRECIZNOST);
              mpfr_init2(alpha[i],PRECIZNOST);
              mpfr_init2(beta[i],PRECIZNOST);
              mpfr_init2(geo_rad[i],PRECIZNOST);

              mpfr_init2(geo_ctr_r[i],PRECIZNOST);
              mpfr_init2(geo_ctr_i[i],PRECIZNOST);
              mpfr_init2(aa_r[i],PRECIZNOST);
              mpfr_init2(aa_i[i],PRECIZNOST);
              mpfr_init2(bb_r[i],PRECIZNOST);
              mpfr_init2(bb_i[i],PRECIZNOST);
        }


        //  ___________________ INITIALIZATION FOR THE EMBEDDING ________________________________

        mpfr_const_pi(alpha[root],MPFR_RNDN);

        mpfr_mul_ui (beta[root], alpha[root], 2, MPFR_RNDN);

        mpfr_sin_cos(aa_i[root],aa_r[root],alpha[root],MPFR_RNDN);
        mpfr_sin_cos(bb_i[root],bb_r[root],beta[root],MPFR_RNDN);


        mpfr_t nula,dva,t1r,t1i,t2r,t2i,t3r,t3i;
        mpfr_inits2 (PRECIZNOST, nula, dva, t1r, t1i, t2r, t2i, t3r, t3i, (mpfr_ptr) 0);
        mpfr_set_d (nula, 0.0, MPFR_RNDN);
        mpfr_set_d (dva, 2.0, MPFR_RNDN);

        mpfr_add(t1r, aa_r[root],bb_r[root],MPFR_RNDN);
        mpfr_add(t1i, aa_i[root],bb_i[root],MPFR_RNDN);
        MPFR_SIGN(t1i)=(MPFR_SIGN(t1i) < 0) ? 1 : -1;  // конјугираме
        c_div(geo_ctr_r[root], geo_ctr_i[root], dva, nula, t1r, t1i);

        mpfr_sub(t1r,geo_ctr_r[root],aa_r[root],MPFR_RNDN);
        mpfr_sub(t1i,geo_ctr_i[root],aa_i[root],MPFR_RNDN);
        mpfr_hypot(geo_rad[root],t1r,t1i,MPFR_RNDN);

        mpfr_set_d (Zr[root], -0.1, MPFR_RNDN);
        mpfr_set_d (Zi[root], -0.1, MPFR_RNDN);


        //  ___________________ LOOP FOR THE EMBEDDING _____________________________

        unsigned long curnode, parent;
        for(i=2;i<=nn;i++){   // zero based are redosled and predecessors
            curnode=redosled[i-1];

            parent=predecessors[curnode-1]; 


            mpfr_set(alpha[curnode],alpha[parent],MPFR_RNDN);
            mpfr_add(t1r,alpha[parent],beta[parent],MPFR_RNDN);
            mpfr_div_ui(beta[curnode],t1r,2.0,MPFR_RNDN);
            mpfr_set(alpha[parent],beta[curnode],MPFR_RNDN);

            mpfr_sin_cos(aa_i[curnode],aa_r[curnode],alpha[curnode],MPFR_RNDN);
            mpfr_sin_cos(bb_i[curnode],bb_r[curnode],beta[curnode],MPFR_RNDN);


            mpfr_add(t1r, aa_r[curnode],bb_r[curnode],MPFR_RNDN);
            mpfr_add(t1i, aa_i[curnode],bb_i[curnode],MPFR_RNDN);
            MPFR_SIGN(t1i)=(MPFR_SIGN(t1i) < 0) ? 1 : -1;  // complex conjugate
            c_div(geo_ctr_r[curnode], geo_ctr_i[curnode], dva, nula, t1r, t1i);

            mpfr_sub(t1r,geo_ctr_r[curnode],aa_r[curnode],MPFR_RNDN);
            mpfr_sub(t1i,geo_ctr_i[curnode],aa_i[curnode],MPFR_RNDN);
            mpfr_hypot(geo_rad[curnode],t1r,t1i,MPFR_RNDN);


            // Implements
            // koord(curnode)=geo_rad(curnode)^2 / (conj(koord(parent)-geo_ctr(curnode))) + geo_ctr(curnode); % complex
            mpfr_mul(t1r,geo_rad[curnode],geo_rad[curnode],MPFR_RNDN);
            mpfr_sub(t2r,Zr[parent],geo_ctr_r[curnode],MPFR_RNDN);
            mpfr_sub(t2i,Zi[parent],geo_ctr_i[curnode],MPFR_RNDN);
            MPFR_SIGN(t2i)=(MPFR_SIGN(t2i) < 0) ? 1 : -1;  // конјугираме
            c_div(t3r,t3i, t1r,nula,t2r,t2i);
            mpfr_add(Zr[curnode],t3r,geo_ctr_r[curnode],MPFR_RNDN);
            mpfr_add(Zi[curnode],t3i,geo_ctr_i[curnode],MPFR_RNDN);

            mpfr_add(t1r,alpha[curnode],beta[curnode],MPFR_RNDN);
            mpfr_div_ui(alpha[curnode],t1r,2.0,MPFR_RNDN);

        }




        //  ___________________ read the topology from file (neighbor list)  _____________________________________________


        FILE *zin1;
        if( (zin1=fopen("spisok-sosedi","r") )==NULL) {return 2;}
        else setvbuf(zin1,NULL,_IOFBF,BUFFER);


        unsigned long **SS;  
        SS = (unsigned long **) malloc ((1+nn)*sizeof(unsigned long *)); 

        unsigned long brsosedi,brjlin=0,j,sosed;

        while (1==fscanf(zin1,"%ld ",&brsosedi)) {
            brjlin++;
         
            SS[brjlin]=(unsigned long *)malloc( (brsosedi+1)*sizeof(unsigned long) );
            SS[brjlin][0]=brsosedi;
            for(j=1; j<=brsosedi; j++) {
                fscanf(zin1,"%ld ",&sosed);
                SS[brjlin][j]=sosed;
            }
        }

        fclose(zin1);



        //  ___________________ forwarding will be done only for some pairs - read list of pairs from file __________________


        if( (zin1=fopen("spisok-parovi","r") )==NULL) {return 3;}
        else setvbuf(zin1,NULL,_IOFBF,BUFFER);

        unsigned long brpar;
        fscanf(zin1,"%ld\n",&brpar);

        unsigned long *sources;
        sources = (unsigned long *) malloc ((1+brpar)*sizeof(unsigned long)); 
        unsigned long *dests;  
        dests = (unsigned long *) malloc ((1+brpar)*sizeof(unsigned long)); 

        for(j=1; j<=brpar; j++) {
            fscanf(zin1,"%ld %ld\n",&sources[j], &dests[j]);
        }


        fclose(zin1);





        //  ___________________ FORWARDING _____________________________________________

        unsigned long ii,jj,msg,index=0;
        mpfr_t currdist;
        mpfr_init2(currdist,PRECIZNOST);

        mpfr_t temp, mindist;

        mpfr_inits2 (PRECIZNOST, temp, mindist, (mpfr_ptr) 0);


        for (ii=1; ii<=brpar; ii++){

            msg=sources[ii];


            hiprast(Zr[msg], Zi[msg], Zr[dests[ii]], Zi[dests[ii]], currdist);



            while (msg != dests[ii]) {

                // FIND A MINIMUM AMONG ALL NEIGHBORS

                mpfr_set_inf (mindist, 1);  


                for( jj=1; jj<=SS[msg][0]; jj++) {  
                    hiprast(Zr[SS[msg][jj]], Zi[SS[msg][jj]], Zr[dests[ii]], Zi[dests[ii]], temp);


                    if (mpfr_less_p(temp,mindist)){
                        mpfr_set(mindist,temp,MPFR_RNDN);
                        index=jj;
                    }
                }


                if(mpfr_less_p(mindist,currdist)) {
                    skoklist_grabl[ii-1] = 1 + skoklist_grabl[ii-1];   

                    msg=SS[msg][index];  

                    mpfr_set(currdist, mindist, MPFR_RNDN);


                } else {
                    vrateno=10;
                    goto higiena;

                }

            }
        }

izlez1:


        //  _______________________ MEMORY CLEANUP ________________________________
higiena:
        mpfr_clears (temp, mindist, (mpfr_ptr) 0);
izlez2:

        free(sources);
        free(dests);

        for(j=1; j<=nn; j++) {
            free(SS[j]);
        }
        free(SS);

        mpfr_clears (nula, dva, t1r, t1i, t2r, t2i, t3r, t3i, (mpfr_ptr) 0);

        for(i=1;i<=nn;i++) {
             mpfr_clear(Zr[i]);
             mpfr_clear(Zi[i]);
             mpfr_clear(alpha[i]);
             mpfr_clear(beta[i]);
             mpfr_clear(geo_rad[i]);

             mpfr_clear(geo_ctr_r[i]);
             mpfr_clear(geo_ctr_i[i]);
             mpfr_clear(aa_r[i]);
             mpfr_clear(aa_i[i]);
             mpfr_clear(bb_r[i]);
             mpfr_clear(bb_i[i]);
        }
        free(Zr); free(Zi); free(alpha); free(beta); free(geo_rad);
        free(geo_ctr_r); free(geo_ctr_i); free(aa_r); free(aa_i);    free(bb_r);    free(bb_i);


        return vrateno;
}
